﻿package 
{
	import flash.events.*;
	import flash.ui.Keyboard;
	import flash.display.MovieClip;
	import flash.geom.*;

	public class Tetris extends MovieClip
	{
		//画面ブロック数
		const FIELD_W:uint = 13;
		const FIELD_H:uint = 25;
		
		//画面ブロックサイズ
		const TIP_W:uint   = 20;
		const TIP_H:uint   = 20;
				
		//確保される配列数
		const ARRAY_W:uint = FIELD_W + 1;
		const ARRAY_H:uint = FIELD_H + 8;
		
		//ブロックの色
		const colorType:Array = [0xFFEEEEEE, 0xFFFF0000, 0xFF00FFFF, 0xFF00FF00, 0xFFAAAAFF, 0xFF0000FF, 0xFFFFAAAA, 0xFF804000];
		
		//ブロックの形
		const blockType:Array = [
			new Block(1, [new Position(0, 0), new Position(0, 0), new Position(0, 0)]), // null
			
			new Block(2, [new Position(0, -1), new Position(0, 1), new Position(0, 2)]), // I
			new Block(4, [new Position(0, -1), new Position(0, 1), new Position(1, 1)]), // L1
			new Block(4, [new Position(0, -1), new Position(0, 1), new Position(-1, 1)]), // L2
			new Block(2, [new Position(0, -1), new Position(1, 0), new Position(1, 1)]), // key1
			new Block(2, [new Position(0, -1), new Position(-1, 0), new Position(-1, 1)]), // key2
			new Block(1, [new Position(0, 1), new Position(1, 0), new Position(1, 1)]), // O
			new Block(4, [new Position(0, -1), new Position(1, 0), new Position( -1, 0)]), // T
			
			//以下変なブロック
			new Block(4, [new Position(0, -1), new Position(1, 0), new Position( -1, 0), new Position( -2, 0)]),
			new Block(2, [new Position(0, -3), new Position(0, -2), new Position(0, -1), new Position(0, 1), new Position(0, 2),new Position(0, 3) ]), 

		];
		

		//画面の状態 (２次元配列
		var board:Array;
		
		//操作中のブロック
		var current:Status = new Status();
		
		//ブロックの落ちる速度
		var blockDownDelay:uint = 6;
		var blockDownDelayCnt:uint = 0;
		
		//画面描画クラス
		var drawer:Drawer;
		
		//キーダウンだけ管理
		var keyStateDown:Boolean = false;

		
		public function Tetris():void
		{
			
			this.x = 10;
			this.y = 10;

			//ボードの生成・初期化
			var i:uint, j:uint;
			
			board = new Array(ARRAY_W);
			
			for (i = 0; i < ARRAY_W; i++) {
				board[i] = new Array(ARRAY_H);
				
				for (j = 0; j < ARRAY_H; j++) {
					
					if (i == 0 || i == ARRAY_W-1 || j == 0) {
						board[i][j] = 1;
					}
					else {
						board[i][j] = 0;
					}
				}
			}
			
			setNextBlock();
			
			//描画クラス
			drawer = new Drawer(this, TIP_W * FIELD_W, TIP_H * FIELD_H);
		}
		
		//初期化
		function init():void
		{
			addEventListener(Event.ENTER_FRAME, update);
			this.stage.addEventListener(KeyboardEvent.KEY_DOWN, onKeyDownEvent);
			this.stage.addEventListener(KeyboardEvent.KEY_UP, onKeyUpEvent);
		}
		//廃棄
		function dispose():void
		{
			removeEventListener(Event.ENTER_FRAME, update);
			this.stage.removeEventListener(KeyboardEvent.KEY_DOWN, onKeyDownEvent);
			this.stage.removeEventListener(KeyboardEvent.KEY_UP, onKeyUpEvent);
			
			drawer.dipose();
		}
		
		
		
		function update(e:Event):void
		{
			
			blockDownDelayCnt++;
		
			//ブロック自動落下
			if (blockDownDelayCnt % blockDownDelay == 0) {
				blockDown();
			}
			
			drawMap();
		}
		
		
		//キーが押された
		function onKeyDownEvent(e:KeyboardEvent):void
		{
			//次の状態
			var n:Status = new Status();
			n.copy(current);
			
			
			if (e.keyCode ==Keyboard.LEFT) {
				n.x--;
			}
			else if (e.keyCode == Keyboard.RIGHT) {
				n.x++;
			}
			
			if (e.keyCode == Keyboard.UP) {
				n.rotate++;
			}
			
			if (n.x != current.x || n.y != current.y || n.rotate != current.rotate) {
				
				deleteBlock(current);
				
				if (putBlock(n)) {
					current = n;
					blockDownDelayCnt = 0;
				}
				else {
					putBlock(current);
				}
				
			}
			
			//一番下まで落下
			else if (e.keyCode == Keyboard.DOWN && keyStateDown == false) {
				putOnBottom();
				keyStateDown = true;
			}
			
		}
		
		//キーが離れた
		function onKeyUpEvent(e:KeyboardEvent):void
		{
			if (e.keyCode == Keyboard.DOWN) {
				keyStateDown = false;
			}
		}
		
		//ブロックを設置
		function putBlock(s:Status, act:Boolean=false):Boolean
		{
			var i:int, j:int;
			
			//置けない
			if (board[s.x][s.y] != 0) {
				return false;
			}
			
			//中心を置く
			if (act) {
				board[s.x][s.y] = s.type;
			}
			
			//周りのブロックを置く
			for (i = 0; i < blockType[s.type].p.length; i++) {
				var dx:int = blockType[s.type].p[i].x;
				var dy:int = blockType[s.type].p[i].y;
				var r:int = s.rotate % blockType[s.type].rotate;
				
				//回転
				for (j = 0;  j < r; j++) {
					var nx:int = dx, ny:int = dy;
					dx = ny; dy = -nx;
				}
				
				//置けない
				if (board[s.x + dx][s.y + dy] != 0) {
					return false;
				}
				
				//ブロックを置く
				if (act) {
					board[s.x + dx][s.y + dy] = s.type;
				}
				
			}
			
			if (!act) {
				putBlock(s, true);
			}
			return true;
		}
		
		//ブロック自動落下
		function blockDown():void
		{
			deleteBlock(current);
			current.y--;
			
			if (!putBlock(current)) {
				current.y++;
				
				putBlock(current);
				
				deleteLine();
				
				setNextBlock();
			}
		}
		
		//次のブロックを生成
		function setNextBlock():void
		{
			current.x = Math.floor(FIELD_W / 2);
			current.y = FIELD_H;
			current.type = 1 + Math.floor(Math.random() * (blockType.length-1));
			current.rotate = Math.floor(Math.random() * 4);
			
			//初期位置に置けないのでガメオベラ
			if (!putBlock(current)) { 
				trace("gameOver");
			}
		}
		
		//ブロック消えるよ！
		function deleteLine():void
		{
			var i:uint, j:uint, x:uint, y:uint;
			var flg:Boolean = true;
			
			for (j = 1; j < FIELD_H; j++) {
				flg = true;
				
				for (i = 1; i <= FIELD_W; i++) {
					//列に空がある
					if (board[i][j] == 0) {
						flg = false;
						break;
					}
				}
				//消える列を発見
				if (flg) {
					for (y = j; y < FIELD_H; y++) {
						for (x = 1; x <= FIELD_W; x++) {
							board[x][y] = board[x][y + 1];
						}
					}
					j--;
				}
			}
			
		}

		//ブロックを一番下に
		function putOnBottom():void
		{
			while (true) {
				deleteBlock(current);
				current.y--;
				
				if(!putBlock(current)) {
					current.y++;
					putBlock(current);
					break;
				}
			}
			blockDownDelayCnt = 0;

		}

		//状態を消します
		function deleteBlock(s:Status):void
		{
			board[s.x][s.y] = 0;
			
			var i:uint, j:uint;
			var dx:int, dy:int;
			var r:int;
			var nx:int, ny:int;
			
			for(i = 0; i < blockType[s.type].p.length; i++) {
				dx = blockType[s.type].p[i].x;
				dy = blockType[s.type].p[i].y;
				r = s.rotate % blockType[s.type].rotate;

				for(j = 0; j < r; j++) {
					nx = dx; ny = dy;
					dx = ny; dy = -nx;
				}
				board[s.x + dx][s.y + dy] = 0;

			}
		}
		
		//画面を書く
		private function drawMap():void
		{
			var i:uint, j:uint;
			var tmpRect:Rectangle = new Rectangle();
			var tmpColor:uint;
			
			
			for (j=1; j<=FIELD_H; j++) {
				for (i = 1; i <= FIELD_W; i++) {
					
					//チップの状態で色を変更
					if (colorType.length > board[i][j]) {
						tmpColor = colorType[board[i][j]];
					}
					else {
						tmpColor = 0xff000000;
					}
					
					//ブロックの四隅を計算
					tmpRect.left   = i * TIP_W;
					tmpRect.right  = tmpRect.left + TIP_W;
					tmpRect.top    = TIP_H * FIELD_H - j * TIP_H;
					tmpRect.bottom = tmpRect.top + TIP_H;
					
					
					drawer.fillRect(tmpRect, tmpColor);
				}
			}
		}
	}
}


class Position
{
	var x:int;
	var y:int;
	
	public function Position(_x:int, _y:int) {
		x = _x;
		y = _y;
	}
}

class Block
{
	var rotate:int;
	var p:Array; //Positionが入る
	
	public function Block(_r:int, _p:Array) {
		rotate= _r;
		p = _p;
	}
}
class Status
{
	var x:int;
	var y:int;
	var type:uint;
	var rotate:int = 0;
	
	function copy(s:Status)
	{
		x = s.x;
		y = s.y;
		type = s.type;
		rotate = s.rotate;
	}
}